//
// dep-info.c
//
// On Vista this program displays the process execution options for all
// processes on the system. On Windows XP the undocumented API we use to get
// this information does not work properly for remote processes, so we show
// the process execution options only for the current process.

#include <stdio.h>

#include <windows.h>
#include <tlhelp32.h>

#define _WIN32_WINNT 0x0500
#include <winternl.h>

BOOL verbose = FALSE;

//
// Returns the address of an exported function in a DLL
//

FARPROC resolve_func(char* dll, char* func)
{
    HMODULE module;
    FARPROC addr;
    
    module = LoadLibrary(dll);
    if (module == NULL) {
        printf("Failed to load DLL %s\n", dll);
        exit(1);
    }

    addr = GetProcAddress(module, func);
    if (addr == NULL) {
        printf("Failed to resolve function %s in DLL %s\n", func, dll);
        exit(1);
    }

    return addr;
}

const DWORD ProcessExecuteFlags = 0x22;

typedef struct {
    unsigned int ExecuteDisable : 1;
    unsigned int ExecuteEnable : 1;
    unsigned int DisableThunkEmulation : 1;
    unsigned int Permanent : 1;
    unsigned int ExecuteDispatchEnable : 1;
    unsigned int ImageDispatchEnable : 1;
    unsigned int DisableExceptionChainValidation : 1;
    unsigned int Spare : 25;
} KEXECUTE_OPTIONS;


//
// Returns the process execution flags
//

BOOL get_process_flags(HANDLE hProcess, KEXECUTE_OPTIONS *flags)
{
    FARPROC NtQueryInformationProcess;
    NTSTATUS status;
    DWORD len;

    NtQueryInformationProcess = resolve_func("ntdll.dll",
        "NtQueryInformationProcess");

    status = NtQueryInformationProcess(hProcess,
                                       ProcessExecuteFlags,
                                       flags,
                                       sizeof(*flags),
                                       &len);

    if (status < 0) {

        // We get this error code for 64-bit processes
        if (status == 0xc000000d)
            return FALSE;

        printf("NtQueryInformationProcess returned error code 0x%x\n", status);
        exit(1);
    }

    if (len != 4) {
        printf("NtQueryInformationProcess returned unexpected length %d\n",
               len);
        exit(1);
    }

    return TRUE;
}

void dump_process_flags(KEXECUTE_OPTIONS flags)
{
    printf(
        "  ExecuteDisable        : %d\n"
        "  ExecuteEnable         : %d\n"
        "  DisableThunkEmulation : %d\n"
        "  Permanent             : %d\n"
        "  ExecuteDispatchEnable : %d\n"
        "  ImageDispatchEnable   : %d\n"
        "  DisableExceptionChainValidation : %d\n",
        flags.ExecuteDisable,
        flags.ExecuteEnable,
        flags.DisableThunkEmulation,
        flags.Permanent,
        flags.ExecuteDispatchEnable,
        flags.ImageDispatchEnable,
        flags.DisableExceptionChainValidation);
}

//
// Iterate through all processes
//

void dump_all_processes()
{
    HANDLE hSnapshot;
    HANDLE hProcess;
    PROCESSENTRY32 pe32;
    DWORD dwPriorityClass;

    // Take a snapshot of all processes in the system

    hSnapshot = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
    if (hSnapshot == INVALID_HANDLE_VALUE) {
        printf("CreateToolhelp32Snapshot returned error code 0x%x\n",
               GetLastError());
        exit(1);
    }

    // Set the size of the structure before using it

    pe32.dwSize = sizeof(pe32);

    // Retrieve information about the first process

    if (!Process32First(hSnapshot, &pe32)) {
        printf("Process32First returned error code 0x%x\n", GetLastError());
        exit(1);
    }

    // Iterate through all processes in the snapshot

    do {
        KEXECUTE_OPTIONS flags;

        HANDLE hProcess = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE,
            pe32.th32ProcessID);

        if (hProcess == NULL) {
            printf("%-5d %-20s [access denied]\n",
                pe32.th32ProcessID, pe32.szExeFile);

            continue;
        }

        if (!get_process_flags(hProcess, &flags))
            printf("%-5d %-20s [query failed]\n",
                pe32.th32ProcessID, pe32.szExeFile);
        else {
            printf("%-5d %-20s 0x%x\n",
                pe32.th32ProcessID, pe32.szExeFile, flags);

            if (verbose)
                dump_process_flags(flags);
        }

        CloseHandle(hProcess);

    } while(Process32Next(hSnapshot, &pe32));

    CloseHandle(hSnapshot);
}

//
// Acquire SeDebugPrivilege
//

void acquire_debug_privilege()
{
    HANDLE hToken;
    TOKEN_PRIVILEGES Priv;

    // Get current thread token

    if  (!OpenThreadToken(GetCurrentThread(),
                          TOKEN_QUERY|TOKEN_ADJUST_PRIVILEGES,
                          FALSE,
                          &hToken))
    {
        // Try to open the current process

        if  (!OpenProcessToken(GetCurrentProcess(),
                              TOKEN_QUERY|TOKEN_ADJUST_PRIVILEGES,
                              &hToken))
        {
            printf("Error in OpenProcessToken\n");
            exit(1);
        }
    }

    // Construct a privilege object with SeDebugPrivilege

    Priv.PrivilegeCount = 1;
    Priv.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

    if (!LookupPrivilegeValue(NULL, SE_DEBUG_NAME, &Priv.Privileges[0].Luid)) {
        printf("Error in LookupPrivilegeValue\n");
        exit(1);
    }

    // Try to enable the privilege

    if(!AdjustTokenPrivileges(hToken, FALSE, &Priv, sizeof(Priv), NULL, NULL)) {
        printf("Error in AdjustTokenPrivileges\n");
        exit(1);
    }
}

int main(int argc, char* argv[])
{
    KEXECUTE_OPTIONS flags;
    OSVERSIONINFO ver;

    if (argc == 2 && strcmp(argv[1], "-v") == 0)
        verbose = TRUE;

    // Get the Windows version

    ver.dwOSVersionInfoSize = sizeof(ver);

    if (GetVersionEx(&ver) == FALSE) {
        printf("Error getting Windows version\n");
        exit(1);
    }

    if (ver.dwMajorVersion <= 5) {

        // Dump only the current process on Windows XP

        printf("Getting information about remote processes works only on Vista.\n\n");

        if (!get_process_flags(GetCurrentProcess(), &flags)) {
            printf("Failed to get process flags\n");
            exit(1);
        }

        printf("Process execution flags of the current process: 0x%x\n", flags);
        dump_process_flags(flags);
    }
    else {

        // Acquire debug privilege

        acquire_debug_privilege();

        // Dump the process flags of all processes

        dump_all_processes();
    }

    return 0;
}
