//
// gs-test.c
//
// The gs-test tool tests the GS function heuristics and variable reordering in
// the Visual C++ compiler. It compiles a number of test cases with different
// local variables and reports whether the functions were protected by GS. Each
// test case is compiled twice - with and without the strict_gs_check pragma.
//  
// To test variable reordering, gs-test gets the addresses of the local
// variables and arguments at runtime and displays them in the order they are
// on the stack.

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>

#include "vars.h"

//
// List of test functions
//

char* gs_tests[] =
{
    #include "tests.list"
};


//
// Returns the address of an exported function in a dll
//

FARPROC resolve_func(char* dll, char* func)
{
    HMODULE module;
    FARPROC addr;
    
    module = LoadLibrary(dll);
    if (module == NULL) {
        printf("Failed to load DLL %s\n", dll);
        exit(1);
    }

    addr = GetProcAddress(module, func);
    if (addr == NULL) {
        printf("Failed to resolve function %s in DLL %s\n", func, dll);
        exit(1);
    }

    return addr;
}


//
// Returns TRUE if the function has a GS cookie
//

BOOL has_gs(FARPROC func)
{
    char* code = (char*)func;

    //
    // Look for the exception handling prologue
    //

    // 55                                push    ebp
    // 8B EC                             mov     ebp, esp
    // 6A FE                             push    0FFFFFFFEh
    // 68 xx xx xx xx                    push    offset unk_100021A8
    // 68 xx xx xx xx                    push    offset _except_handler4
    // 64 A1 00 00 00 00                 mov     eax, large fs:0
    // 50                                push    eax

    if (memcmp(code, "\x55\x8B\xEC\x6A\xFE\x68", 6) == 0 &&
        code[10] == '\x68' &&
        memcmp(&code[15], "\x64\xA1\x00\x00\x00\x00\x50", 7) == 0)
    {
        code += 22;

        //
        // Skip over instructions that allocate space on the stack
        //

        // 83 EC 08                 sub     esp, 8
        // B8 xx xx xx xx           mov     eax, dword
        // E8 xx xx xx xx           call    _chkstk

        if (code[0] == '\x83' && code[1] == '\xEC' && code[2] == '\x08' &&
                 code[3] == '\xB8' && code[8] == '\xE8')
            code += 13;

        // 83 EC xx                 sub     esp, byte

        else if (code[0] == '\x83' && code[1] == '\xEC')
            code += 3;

        // 81 EC xx xx xx xx        sub     esp, dword

        else if (code[0] == '\x81' && code[1] == '\xEC')
            code += 6;

        // 83 C4 xx                 add     esp, -byte

        else if (code[0] == '\x83' && code[1] == '\xC4')
            code += 3;

        // 81 C4 xx xx xx xx        sub     esp, -dword

        else if (code[0] == '\x81' && code[1] == '\xC4')
            code += 6;

        //
        // Look for the security cookie code
        //

        // A1 xx xx xx xx                    mov     eax, __security_cookie
        // 31 45 F8                          xor     [ebp+scope_table], eax
        // 33 C5                             xor     eax, ebp
        // 89 45 E4                          mov     [ebp+gs_cookie], eax

        if (code[0] == '\xA1' &&
            memcmp(&code[5], "\x31\x45\xF8\x33\xC5\x89\x45\xE4", 8) == 0)
        {
            return TRUE;
        }

        // No stack cookie found

        return FALSE;
    }

    //
    // No exception handling code found
    //

    //
    // Skip over the instructions that set up the base pointer
    //

    // 55                       push    ebp
    // 8B EC                    mov     ebp, esp

    if (memcmp(code, "\x55\x8B\xEC", 3) == 0)
        code += 3;

    //
    // Skip over instructions that allocate space on the stack
    //

    // 51                       push    ecx

    if (code[0] == '\x51')
        code += 1;

    // 83 EC xx                 sub     esp, byte

    else if (code[0] == '\x83' && code[1] == '\xEC')
        code += 3;

    // 81 EC xx xx xx xx        sub     esp, dword

    else if (code[0] == '\x81' && code[1] == '\xEC')
        code += 6;

    // 83 C4 xx                 add     esp, -byte

    else if (code[0] == '\x83' && code[1] == '\xC4')
        code += 3;

    // 81 C4 xx xx xx xx        sub     esp, -dword

    else if (code[0] == '\x81' && code[1] == '\xC4')
        code += 6;

    // B8 xx xx xx xx           mov     eax, dword
    // E8 xx xx xx xx           call    _chkstk

    else if ((code[0] == '\xB8') && (code[5] == '\xE8'))
        code += 10;

    //
    // Look for the security cookie code
    //

    // A1 xx xx xx xx           mov     eax, __security_cookie
    // 33 C4                    xor     eax, esp

    if (code[0] == '\xA1' && code[5] == '\x33' && code[6] == '\xC4')
        return TRUE;
    
    // A1 xx xx xx xx           mov     eax, __security_cookie
    // 33 C5                    xor     eax, ebp

    if (code[0] == '\xA1' && code[5] == '\x33' && code[6] == '\xC5')
        return TRUE;

    // No stack cookie found

    return FALSE;
}


//
// Compare function for sorting arrays of struct var
//

int __cdecl var_compare(const void* a, const void* b)
{
    const struct var* var1 = a;
    const struct var* var2 = b;

    if (var1->addr < var2->addr)
        return -1;
    else if (var1->addr > var2->addr)
        return 1;
    else
        return 0;
}


//
// Dumps all variables from the reorder test functions
//

typedef vars_t* (*test_reorder_fn)();

void dump_variables(FARPROC test_nogs, FARPROC test_gs, FARPROC test_gs_strict)
{
    unsigned int count, i;

    vars_t* vars_source    = ((test_reorder_fn)test_nogs)();
    vars_t* vars_nogs      = ((test_reorder_fn)test_nogs)();
    vars_t* vars_gs        = ((test_reorder_fn)test_gs)();
    vars_t* vars_gs_strict = ((test_reorder_fn)test_gs_strict)();

    count = vars_source->count;

    if (vars_nogs->count != count ||
        vars_gs->count != count ||
        vars_gs_strict->count != count)
    {
        printf("Inconsistent variable count\n");
        exit(1);
    }


    // Sort the variables based on their address on the stack, except for the
    // vars_source array

    qsort(&vars_nogs->var, count, sizeof(struct var), var_compare);
    qsort(&vars_gs->var, count, sizeof(struct var), var_compare);
    qsort(&vars_gs_strict->var, count, sizeof(struct var), var_compare);

    // Print all variables

    for (i = 0; i < count; i++) {
        printf("    %-19s %-18s %-18s %s\n",
            vars_source->var[i].name, vars_nogs->var[i].name,
            vars_gs->var[i].name, vars_gs_strict->var[i].name);
    }

    // Free the arrays

    vars_free(vars_source);
    vars_free(vars_nogs);
    vars_free(vars_gs);
    vars_free(vars_gs_strict);
}


//
// Run all tests
//

int main()
{
    int i;

    //
    // GS heuristics
    //

    printf("Testing GS heuristics:\n\n");

    printf("  %-40s %-12s %s\n\n", "test case", "default GS", "strict GS");

    for (i = 0; gs_tests[i] != NULL; i++) {
        printf("  %-40s %-12s %s\n", gs_tests[i],
            has_gs(resolve_func("tests-gs.dll", gs_tests[i])) ?
                "GS" : "missing",

            has_gs(resolve_func("tests-gs-strict.dll", gs_tests[i])) ?
                "GS" : "missing"
        );
    }

    //
    // Variable reordering
    //

    printf("\nTesting variable reordering:\n\n");

    for (i = 0; gs_tests[i] != NULL; i++) {
        if (!strncmp(gs_tests[i], "test_reorder_", sizeof("test_reorder_")-1))
        {
            printf("  %s\n\n    disabled : %s\n    default  : %s\n    strict   : %s\n\n",
                gs_tests[i],

                has_gs(resolve_func("tests-nogs.dll", gs_tests[i])) ?
                    "GS" : "missing",

                has_gs(resolve_func("tests-gs.dll", gs_tests[i])) ?
                    "GS" : "missing",

                has_gs(resolve_func("tests-gs-strict.dll", gs_tests[i])) ?
                    "GS" : "missing"
            );

            printf("    %-19s %-18s %-18s %s\n\n",
                "source code order", "disabled GS", "default GS", "strict GS");

            dump_variables(resolve_func("tests-nogs.dll", gs_tests[i]),
                           resolve_func("tests-gs.dll", gs_tests[i]),
                           resolve_func("tests-gs-strict.dll", gs_tests[i]));

            printf("\n");
        }
    }

    return 0;
}
